{-|
Module      : Gargantext.Core.Worker
Description : Asynchronous worker logic
Copyright   : (c) CNRS, 2024
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

-}

{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE UndecidableInstances #-}
{-# OPTIONS_GHC -Wno-orphans      #-}  -- orphan HasNodeError IOException


module Gargantext.Core.Worker where


import Async.Worker.Broker.PGMQ (PGMQBroker)
import Async.Worker.Broker.Types (BrokerMessage, toA, getMessage, messageId)
import Async.Worker qualified as Worker
import Async.Worker.Types qualified as Worker
import Async.Worker.Types (HasWorkerBroker)
import Data.Text qualified as T
import Gargantext.API.Admin.Auth (forgotUserPassword)
import Gargantext.API.Admin.Auth.Types (ForgotPasswordAsyncParams(..))
import Gargantext.API.Node.Corpus.New (addToCorpusWithForm, addToCorpusWithQuery)
import Gargantext.API.Node.New (postNode')
import Gargantext.API.Node.Types (WithQuery(..))
import Gargantext.Core.Config (hasConfig, gc_jobs)
import Gargantext.Core.Config.Types (jc_max_docs_scrapers)
import Gargantext.Core.Config.Worker (WorkerDefinition(..))
import Gargantext.Core.Notifications.CentralExchange.Types qualified as CE
import Gargantext.Core.Worker.Broker (initBrokerWithDBCreate)
import Gargantext.Core.Worker.Env
import Gargantext.Core.Worker.Jobs.Types (Job(..))
import Gargantext.Core.Worker.Types (JobInfo(..))
import Gargantext.Database.Admin.Types.Node (NodeId(..))
import Gargantext.Database.Query.Table.User (getUsersWithEmail)
import Gargantext.Prelude
import Gargantext.System.Logging ( logLocM, LogLevel(..) )



initWorkerState :: (HasWorkerBroker PGMQBroker Job)
                => WorkerEnv
                -> WorkerDefinition
                -> IO (Worker.State PGMQBroker Job)
initWorkerState env (WorkerDefinition { .. }) = do
  let gargConfig = env ^. hasConfig
  broker <- initBrokerWithDBCreate gargConfig

  pure $ Worker.State { broker
                      , queueName = _wdQueue
                      , name = T.unpack _wdName
                      , performAction = performAction env
                      , onMessageReceived = Just $ markJobStarted env
                      , onJobFinish = Nothing
                      , onJobTimeout = Just $ \_s bm -> putStrLn ("on job timeout: " <> show (toA $ getMessage bm) :: Text)
                      , onJobError = Nothing
                      , onWorkerKilledSafely = Nothing }

markJobStarted :: (HasWorkerBroker PGMQBroker Job)
               => WorkerEnv
               -> Worker.State PGMQBroker Job
               -> BrokerMessage PGMQBroker (Worker.Job Job)
               -> IO ()
markJobStarted env (Worker.State { name }) bm = do
  let j = toA $ getMessage bm
  putStrLn $ "[" <> name <> "] starting job: " <> show j
  let ji = JobInfo { _ji_message_id = messageId bm }
  case Worker.job j of
    AddCorpusWithQuery { _acq_args = WithQuery { _wq_node_id } } -> do
      runWorkerMonad env $ CE.ce_notify $ CE.WorkerJobStarted (UnsafeMkNodeId _wq_node_id) ji
    _ -> pure ()


-- | Spawn a worker with PGMQ broker
-- TODO:
--  - reduce size of DB pool
--  - progress report via notifications
--  - I think there is no point to save job result, as usually there is none (we have side-effects only)
--  - replace Servant.Job to use workers instead of garg API threads
withPGMQWorker :: (HasWorkerBroker PGMQBroker Job)
                => WorkerEnv
                -> WorkerDefinition
                -> (Async () -> Worker.State PGMQBroker Job -> IO ())
                -> IO ()
withPGMQWorker env wd cb = do
  state' <- initWorkerState env wd

  withAsync (Worker.run state') (\a -> cb a state')

withPGMQWorkerSingle :: (HasWorkerBroker PGMQBroker Job)
                     => WorkerEnv
                     -> WorkerDefinition
                     -> (Async () -> Worker.State PGMQBroker Job -> IO ())
                     -> IO ()
withPGMQWorkerSingle env wd cb = do
  state' <- initWorkerState env wd

  withAsync (Worker.runSingle state') (\a -> cb a state')


-- | How the worker should process jobs
performAction :: (HasWorkerBroker PGMQBroker Job)
              => WorkerEnv
              -> Worker.State PGMQBroker Job
              -> BrokerMessage PGMQBroker (Worker.Job Job)
              -> IO ()
performAction env _state bm = do
  let job' = toA $ getMessage bm
  let ji = JobInfo { _ji_message_id = messageId bm }
  let jh = WorkerJobHandle { _w_job_info = ji }
  case Worker.job job' of
    Ping -> runWorkerMonad env $ do
      $(logLocM) DEBUG "[performAction] ping"
    AddCorpusFormAsync { .. } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "[performAction] add corpus form"
      addToCorpusWithForm _acf_user _acf_cid _acf_args jh
    AddCorpusWithQuery { .. } -> runWorkerMonad env $ do
      $(logLocM) DEBUG "[performAction] add corpus with query"
      let limit = Just $ fromIntegral $ env ^. hasConfig . gc_jobs . jc_max_docs_scrapers
      addToCorpusWithQuery _acq_user _acq_cid _acq_args limit jh
    ForgotPasswordAsync { _fpa_args = ForgotPasswordAsyncParams { email } } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "[performAction] forgot password: " <> email
      us <- getUsersWithEmail (T.toLower email)
      case us of
        [u] -> forgotUserPassword u
        _ -> pure ()
    NewNodeAsync { .. } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "[performAction] new node async "
      void $ postNode' _nna_authenticatedUser _nna_node_id _nna_postNode
    GargJob { _gj_garg_job } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "Garg job: " <> show _gj_garg_job <> " (handling of this job is still not implemented!)"
      return ()
